/*****************************************************************************/
/*                                                                           */
/* injenv.cpp: Source file for environment variable changer                  */
/*                                                  for another proccess     */
/*                                                                           */
/*             Copyright (c) Yak! 2007-2008 All rights reserved.             */
/*                                                                           */
/*   $URL: https://yak.myhome.cx/repos/source/afxscrpt/trunk/injenv.cpp $    */
/*  $Date: 2011-10-30 20:32:44 +0900 (日, 30 10月 2011) $                    */
/*   $Rev: 1127 $                                                             */
/*                                                                           */
/*****************************************************************************/

#include <windows.h>
#include <algorithm>
#include <limits>

const char* KEY_INJENVVARARG = "yak.myhome.cx::InjectEnvironmentVairable::Arg";
const char* KEY_INJENVVARMES = "yak.myhome.cx::InjectEnvironmentVairable::Message";
const char* KEY_INJENVVARCTX = "yak.myhome.cx::InjectEnvironmentVairable::Context";

const char* KEY_INJCMDARG = "yak.myhome.cx::InjectCommand::Arg";
const char* KEY_INJCMDMES = "yak.myhome.cx::InjectCommand::Message";

const char* KEY_INJMESARG = "yak.myhome.cx::InjectMessage::Arg";
const char* KEY_INJMESMES = "yak.myhome.cx::InjectMessage::Message";

const char* KEY_INJKEYMES = "yak.myhome.cx::InjectKey::Message";

// Prolog of eval function
const char MARKER_CMDLA[12] = {
	'\x55', '\x8b', '\xec', '\x6a', '\x00', '\x53', '\x8b', '\xd8',
	'\x33', '\xc0', '\x55', '\x68'
};
// "ss : " NOTE that it includes terminator
const char MARKER_CMD[24] = {
	'\x8E', '\xC0', '\x8D', '\x73', '\x8E', '\xB8', '\x94', '\x73',
	'\x20', '\x3A', '\x20', '\x00',

	'\x55', '\x8b', '\xec', '\x6a', '\x00', '\x53', '\x8b', '\xd8',
	'\x33', '\xc0', '\x55', '\x68'
};
const char MARKER_CMDW[28] = {
	'\x9f', '\x5b', '\x4c', '\x88', '\x31', '\x59', '\x57', '\x65',
	'\x20', '\x00', '\x3A', '\x00', '\x20', '\x00', '\x00', '\x00',

	'\x55', '\x8b', '\xec', '\x6a', '\x00', '\x53', '\x8b', '\xd8',
	'\x33', '\xc0', '\x55', '\x68'
};

const char MARKER_MES[20] = {
	'\x55', '\x8b', '\xec', '\x83', '\xc4', '\xf8', '\x53', '\x56',
	'\x57', '\x33', '\xd2', '\x89', '\x55', '\xf8', '\x8b', '\xf8',
	'\x33', '\xc0', '\x55', '\x68'
};
const char MARKER_MESW[20] = {
	'\x55', '\x8b', '\xec', '\x83', '\xc4', '\xf8', '\x53', '\x56',
	'\x57', '\x33', '\xd2', '\x89', '\x55', '\xf8', '\x89', '\x45',
	'\xfc', '\x8b', '\x45', '\xfc'
};
const char MARKER_MESW1[32] = {
	'\xee', '\x4e', '\xf3', '\x60', '\xd5', '\x30', '\xa9', '\x30',
	'\xeb', '\x30', '\xc0', '\x30', '\x6e', '\x30', '\xad', '\x8a',
	'\xbc', '\x8f', '\x92', '\x30', '\x2d', '\x4e', '\xad', '\x65',
	'\x57', '\x30', '\x7e', '\x30', '\x57', '\x30', '\x5f', '\x30'
};
char MARKER_MESW2[6] = {
	'\xb8', '\x00', '\x00', '\x00', '\x00', '\xe8'
};

const char PREFIX[8] = {
	'\xb0', '\x04', '\x02', '\x00', '\xFF', '\xFF', '\xFF', '\xFF'

};

//////////////////////////////////////////////////////////////////////////////
///
/// Helper class for shared memory by file mapping
///
//////////////////////////////////////////////////////////////////////////////
class Str
{
public:
	Str(void *p, bool fSrcUnicode);
	Str(LPSTR p);
	Str(LPWSTR p);
	~Str();
	operator LPSTR();
	operator LPWSTR();
private:
	LPSTR m_pStrA;
	bool fAllocA;
	LPWSTR m_pStrW;
	bool fAllocW;
};

Str::Str(void *p, bool fSrcUnicode) : m_pStrA(0), fAllocA(false), m_pStrW(0), fAllocW(false)
{
	if(fSrcUnicode) {
		m_pStrW = static_cast<LPWSTR>(p);
	} else {
		m_pStrA = static_cast<LPSTR>(p);
	}
}

Str::Str(LPSTR p) : m_pStrA(p), fAllocA(false), m_pStrW(0), fAllocW(false)
{
}

Str::Str(LPWSTR p) : m_pStrA(0), fAllocA(false), m_pStrW(p), fAllocW(false)
{
}

Str::operator LPSTR()
{
	if(!m_pStrA && m_pStrW) {
		int nLen = WideCharToMultiByte(CP_ACP, 0, m_pStrW, -1, 0, 0, 0, 0);
		m_pStrA = new char[nLen];
		fAllocA = true;
		WideCharToMultiByte(CP_ACP, 0, m_pStrW, -1, m_pStrA, nLen, 0, 0);
	}
	return m_pStrA;
}

Str::operator LPWSTR()
{
	if(!m_pStrW && m_pStrA) {
		int nLen = MultiByteToWideChar(CP_ACP, 0, m_pStrA, -1, 0, 0);
		m_pStrW = new WCHAR[nLen];
		fAllocW = true;
		MultiByteToWideChar(CP_ACP, 0, m_pStrA, -1, m_pStrW, nLen);
	}
	return m_pStrW;
}

Str::~Str()
{
	if(fAllocA) delete[] m_pStrA;
	if(fAllocW) delete[] m_pStrW;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Thin wrapper class for shared memory by file mapping
///
//////////////////////////////////////////////////////////////////////////////
class SharedMem
{
public:
	explicit SharedMem(LPCTSTR lpName, size_t size);
	~SharedMem();
	void* GetPointer(void) { return m_pBuffer; }
private:
	HANDLE m_hMapping;
	void* m_pBuffer;
};

SharedMem::SharedMem(LPCTSTR lpName, size_t size) :
       m_hMapping(CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, size, lpName)), m_pBuffer(NULL)
{
	if(m_hMapping == NULL || m_hMapping == INVALID_HANDLE_VALUE) {
		throw 1;
	}
	m_pBuffer = MapViewOfFile(m_hMapping, FILE_MAP_ALL_ACCESS, 0, 0, size);
	if(!m_pBuffer) throw 1;
}

SharedMem::~SharedMem()
{
	UnmapViewOfFile(m_pBuffer);
	CloseHandle(m_hMapping);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Shared data
///
//////////////////////////////////////////////////////////////////////////////
const int MAX_HWND = std::numeric_limits<BYTE>::digits;
struct SharedData
{
	HHOOK hHook;
	BYTE bFlag;
	HWND hwnd[MAX_HWND];
};

//////////////////////////////////////////////////////////////////////////////
///
/// Global variables
///
//////////////////////////////////////////////////////////////////////////////
SharedMem *g_pSharedMem;
HINSTANCE g_hInstance;
UINT g_uInjEnvMessage;
UINT g_uInjCmdMessage;
UINT g_uInjMesMessage;
UINT g_uInjKeyMessage;

//////////////////////////////////////////////////////////////////////////////
///
/// Helper functions for command / message / key injection
///
//////////////////////////////////////////////////////////////////////////////

// In gcc, __fastcall does not mean using EAX for argument passing.
// This is the same behavior as MSVC.
// However, Borland/Codegear uses different calling convention.
typedef int (__attribute__((regparm(1))) *EVALFUNC)(LPSTR);
typedef int (__attribute__((regparm(1))) *EVALFUNCW)(LPWSTR);

const BYTE* const cstart = reinterpret_cast<BYTE*>(0x00401000);
static const BYTE* GetCodeEnd()
{
	static const BYTE* cend = 0;
	if(!cend) {
		MEMORY_BASIC_INFORMATION mbi;
		VirtualQuery(cstart, &mbi, sizeof(mbi));
		cend = cstart + mbi.RegionSize;
	}
	return cend;
}

static EVALFUNC GetEvalFunc()
{
	static EVALFUNC routine = 0;
	if(!routine) {
		const BYTE* cend   = GetCodeEnd();
		const BYTE* tstart = static_cast<const BYTE*>(static_cast<const void*>(MARKER_CMD));
		const BYTE* tend   = static_cast<const BYTE*>(static_cast<const void*>(MARKER_CMD + sizeof(MARKER_CMD)));
		const BYTE* result = std::search(cstart, cend, tstart, tend);
		if(result != cend) {
			OutputDebugString("EvalFunc Found");
			routine = reinterpret_cast<EVALFUNC>(reinterpret_cast<DWORD_PTR>(result + sizeof(MARKER_CMD) - sizeof(MARKER_CMDLA)));
		}
	}
	return routine;
}

static EVALFUNCW GetEvalFuncW()
{
	static EVALFUNCW routine = 0;
	if(!routine) {
		const BYTE* cend   = GetCodeEnd();
		const BYTE* tstart = static_cast<const BYTE*>(static_cast<const void*>(MARKER_CMDW));
		const BYTE* tend   = static_cast<const BYTE*>(static_cast<const void*>(MARKER_CMDW + sizeof(MARKER_CMDW)));
		const BYTE* result = std::search(cstart, cend, tstart, tend);
		if(result != cend) {
			OutputDebugString("EvalFuncW Found");
			routine = reinterpret_cast<EVALFUNCW>(reinterpret_cast<DWORD_PTR>(result + sizeof(MARKER_CMDW) - sizeof(MARKER_CMDLA)));
		}
	}
	return routine;
}

static EVALFUNC GetMessageFunc()
{
	static EVALFUNC routine = 0;
	if(!routine) {
		const BYTE* cend   = GetCodeEnd();
		const BYTE* tstart = static_cast<const BYTE*>(static_cast<const void*>(MARKER_MES));
		const BYTE* tend   = static_cast<const BYTE*>(static_cast<const void*>(MARKER_MES + sizeof(MARKER_MES)));
		const BYTE* result = std::search(cstart, cend, tstart, tend);
		if(result != cend) {
			OutputDebugString("MessageFunc Found");
			routine = reinterpret_cast<EVALFUNC>(reinterpret_cast<DWORD_PTR>(result));
		}
	}
	return routine;
}

static EVALFUNCW GetMessageFuncW()
{
	static EVALFUNCW routine = 0;
	if(!routine) {
		const BYTE* cend   = GetCodeEnd();
		const BYTE* tstart = static_cast<const BYTE*>(static_cast<const void*>(MARKER_MESW1));
		const BYTE* tend   = static_cast<const BYTE*>(static_cast<const void*>(MARKER_MESW1 + sizeof(MARKER_MESW1)));
		const BYTE* result = std::search(cstart, cend, tstart, tend);
		if(result != cend) {
			tstart = static_cast<const BYTE*>(static_cast<const void*>(MARKER_MESW2));
			tend   = static_cast<const BYTE*>(static_cast<const void*>(MARKER_MESW2 + sizeof(MARKER_MESW2)));
			char* tresult = static_cast<char*>(static_cast<void*>(&result));
			std::copy(tresult, tresult + 4, MARKER_MESW2 + 1);
			result = std::search(cstart, cend, tstart, tend);
			if(result != cend) {
				OutputDebugString("MessageFuncW Found");
				result += *static_cast<const DWORD*>(static_cast<const void*>(result + 6)) + 5 + 5;
				routine = reinterpret_cast<EVALFUNCW>(reinterpret_cast<DWORD_PTR>(result));
			}
		}
	}
	return routine;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Hook procedure for environment variable change and command injection
///
//////////////////////////////////////////////////////////////////////////////
static void ProcessInjEnvMessage(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	try {
		OutputDebugString("HookCalled by InjEnv");
		SharedMem smArg(KEY_INJENVVARARG, wParam);
		if(lParam) { // Unicode
			LPWSTR pFirst = static_cast<LPWSTR>(smArg.GetPointer());
			LPWSTR pSecond = pFirst;
			while(*pSecond++) ;
			OutputDebugStringW(pFirst);
			OutputDebugStringW(pSecond);
			if(SetEnvironmentVariableW(pFirst, pSecond)) {
				*pFirst = 0;
			}
		} else {
			LPSTR pFirst = static_cast<LPSTR>(smArg.GetPointer());
			LPSTR pSecond = pFirst;
			while(*pSecond++) ;
			OutputDebugStringA(pFirst);
			OutputDebugStringA(pSecond);
			if(SetEnvironmentVariableA(pFirst, pSecond)) {
				*pFirst = 0;
			}
		}
	} catch(...) {
		OutputDebugString("ErrorOccurred by InjEnv");
	}
}

static void ProcessInjCmdMessage(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	try {
		OutputDebugString("HookCalled by InjCmd");
		SharedMem smArg(KEY_INJCMDARG, wParam);
		if(lParam) { // Unicode
			LPSTR pFirst = static_cast<LPSTR>(smArg.GetPointer());
			LPWSTR pStart = static_cast<LPWSTR>(static_cast<LPVOID>(pFirst + sizeof(PREFIX) + sizeof(int)));
			OutputDebugStringW(pStart);
			EVALFUNCW fEval = GetEvalFuncW();
			if(fEval) {
				fEval(pStart);
			} else {
				OutputDebugString("No eval function for unicode found.");
			}
		} else {
			LPSTR pFirst = static_cast<LPSTR>(smArg.GetPointer());
			OutputDebugStringA(pFirst + sizeof(int));
			EVALFUNC fEval = GetEvalFunc();
			if(fEval) {
				fEval(pFirst + sizeof(int));
			} else {
				OutputDebugString("No eval function found.");
			}
		}
	} catch(...) {
		OutputDebugString("ErrorOccurred by InjCmd");
	}
}

static void ProcessInjMesMessage(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	try {
		OutputDebugString("HookCalled by InjMes");
		SharedMem smArg(KEY_INJMESARG, wParam);
		if(lParam) { // Unicode
			LPSTR pFirst = static_cast<LPSTR>(smArg.GetPointer());
			LPWSTR pStart = static_cast<LPWSTR>(static_cast<LPVOID>(pFirst + sizeof(PREFIX) + sizeof(int)));
			OutputDebugStringW(pStart);
			EVALFUNCW fEval = GetMessageFuncW();
			if(fEval) {
				fEval(pStart);
			} else {
				OutputDebugString("No message function found.");
			}
		} else {
			char* pFirst = static_cast<char*>(smArg.GetPointer());
			OutputDebugString(pFirst + sizeof(int) * 2);
			EVALFUNC fEval = GetMessageFunc();
			if(fEval) {
				fEval(pFirst + sizeof(int) * 2);
			} else {
				OutputDebugString("No message function found.");
			}
		}
	} catch(...) {
		OutputDebugString("ErrorOccurred by InjMes");
	}
}

static void ProcessInjKeyMessage(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	hwnd = GetFocus();
	OutputDebugString("HookCalled by InjKey");
	int nModifier = wParam / 1000;
	int nKey = wParam % 1000;
	BYTE table[256], table_backup[256];
	GetKeyboardState(table);
	CopyMemory(table_backup, table, sizeof(table));

	if(nModifier & 1) {
		table[VK_SHIFT] = 0x80;
		SetKeyboardState(table);
	}
	if(nModifier & 2) {
		table[VK_CONTROL] = 0x80;
		SetKeyboardState(table);
	}

	if(nModifier & 4) {
		SendMessage(hwnd, WM_SYSKEYDOWN, nKey,       0x20000001);
		SendMessage(hwnd, WM_SYSKEYUP,   nKey,       0xE0000001);
	} else {
		SendMessage(hwnd, WM_KEYDOWN,    nKey,       0x001E0001);
		SendMessage(hwnd, WM_KEYUP,      nKey,       0xC01E0001);
	}

	if(nModifier & 1 || nModifier & 2)
		SetKeyboardState(table_backup);
}

static HWND AfxWindowFromTFileBox(HWND hwnd)
{
	char buf[2048];
	GetClassName(hwnd, buf, sizeof(buf));
	if(lstrcmp(buf, "TFileBox")) return NULL;
	hwnd = GetParent(hwnd);
	if(!hwnd) return NULL;
	hwnd = GetParent(hwnd);
	if(!hwnd) return NULL;
	GetClassName(hwnd, buf, sizeof(buf));
	if(lstrcmp(buf, "TAfxForm") && lstrcmp(buf, "TAfxWForm")) return NULL;
	return hwnd;
}

static void ProcessGetCaretIndex(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	HWND hwndAfx = AfxWindowFromTFileBox(hwnd);
	if(!hwndAfx) return;

	static HWND hwndPrev;
	if(hwndPrev == hwnd) return;
	hwndPrev = hwnd;

	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	for(int i = 0, mask = 1; i < MAX_HWND; ++i, mask <<= 1) {
		if(pSharedData->bFlag & mask) {
			PostMessage(pSharedData->hwnd[i], WM_APP, reinterpret_cast<WPARAM>(hwndAfx), 0);
		}
	}
}

static void ProcessSetCaretIndex(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	HWND hwndAfx = AfxWindowFromTFileBox(hwnd);
	if(!hwndAfx) return;

	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	for(int i = 0, mask = 1; i < MAX_HWND; ++i, mask <<= 1) {
		if(pSharedData->bFlag & mask) {
			PostMessage(pSharedData->hwnd[i], WM_APP, reinterpret_cast<WPARAM>(hwndAfx), 0);
		}
	}
}

static bool IsAfxWindow(HWND hwnd)
{
	char buf[2048];
	GetClassName(hwnd, buf, sizeof(buf));
	if(lstrcmp(buf, "TAfxForm") && lstrcmp(buf, "TAfxWForm")) return false;
	return true;
}

static void ProcessMoveMessage(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	for(int i = 0, mask = 1; i < MAX_HWND; ++i, mask <<= 1) {
		if(pSharedData->bFlag & mask) {
			PostMessage(pSharedData->hwnd[i], WM_APP+1, reinterpret_cast<WPARAM>(hwnd), lParam);
		}
	}
}

static bool IsAfxWWindow(HWND hwnd)
{
	char buf[2048];
	GetClassName(hwnd, buf, sizeof(buf));
	if(lstrcmp(buf, "TAfxWForm")) return false;
	return true;
}

static bool IsAfxAWindow(HWND hwnd)
{
	char buf[2048];
	GetClassName(hwnd, buf, sizeof(buf));
	if(lstrcmp(buf, "TAfxForm")) return false;
	return true;
}

static bool IsTApplication(HWND hwnd)
{
	char buf[2048];
	GetClassName(hwnd, buf, sizeof(buf));
	if(lstrcmp(buf, "TApplication")) return false;
	return true;
}

struct AfxWindowFromTApplicationData
{
	HWND hwndResult;
	DWORD dwProcessId;
};

static BOOL CALLBACK AfxWindowFromTApplicationProc(HWND hwnd, LPARAM lParam)
{
	if(!IsAfxWindow(hwnd)) return TRUE;

	AfxWindowFromTApplicationData *pData = reinterpret_cast<AfxWindowFromTApplicationData *>(lParam);

	DWORD dwProcessId;
	GetWindowThreadProcessId(hwnd, &dwProcessId);
	if(pData->dwProcessId != dwProcessId) return TRUE;

	pData->hwndResult = hwnd;
	return FALSE;
}

static HWND AfxWindowFromTApplication(HWND hwnd)
{
	if(!IsTApplication(hwnd)) return NULL;

	DWORD dwProcessId;
	GetWindowThreadProcessId(hwnd, &dwProcessId);

	AfxWindowFromTApplicationData data = {
		NULL,
		dwProcessId
	};
	EnumWindows(AfxWindowFromTApplicationProc, reinterpret_cast<LPARAM>(&data));

	return data.hwndResult;
}

static BOOL CALLBACK TApplicationFromAfxWindowProc(HWND hwnd, LPARAM lParam)
{
	if(!IsTApplication(hwnd)) return TRUE;

	AfxWindowFromTApplicationData *pData = reinterpret_cast<AfxWindowFromTApplicationData *>(lParam);

	DWORD dwProcessId;
	GetWindowThreadProcessId(hwnd, &dwProcessId);
	if(pData->dwProcessId != dwProcessId) return TRUE;

	pData->hwndResult = hwnd;
	return FALSE;
}

static HWND TApplicationFromAfxWindow(HWND hwnd)
{
	if(!IsAfxWindow(hwnd)) return NULL;

	DWORD dwProcessId;
	GetWindowThreadProcessId(hwnd, &dwProcessId);

	AfxWindowFromTApplicationData data = {
		NULL,
		dwProcessId
	};
	EnumWindows(TApplicationFromAfxWindowProc, reinterpret_cast<LPARAM>(&data));

	return data.hwndResult;
}

static bool IsMainFormOnTaskbar(HWND hwnd)
{
	if(!IsTApplication(hwnd)) {
OutputDebugString("IsMainFormOnTaskbar: Precondition failed.");
		return false;
	}

	if(GetWindowTextLength(hwnd)) {
		return false;
	} else {
		return true;
	}
}

static ATOM CreateMessageArgAtom(WPARAM wParam, LPARAM lParam)
{
	char buf[256];
	wsprintf(buf, "0x%08X|0x%08X", wParam, lParam);
	return GlobalAddAtom(buf);
}

static void ProcessSizeMessage(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	if(!IsAfxWindow(hwnd)) {
		if(wParam != SIZE_MINIMIZED) return;
		if(!IsTApplication(hwnd) || IsMainFormOnTaskbar(hwnd)) return;
		hwnd = AfxWindowFromTApplication(hwnd);
		if(!hwnd) return;
	}
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	for(int i = 0, mask = 1; i < MAX_HWND; ++i, mask <<= 1) {
		if(pSharedData->bFlag & mask) {
			PostMessage(pSharedData->hwnd[i], WM_APP+2, reinterpret_cast<WPARAM>(hwnd), CreateMessageArgAtom(wParam, lParam));
		}
	}
}

static void ProcessActivateMessage(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	for(int i = 0, mask = 1; i < MAX_HWND; ++i, mask <<= 1) {
		if(pSharedData->bFlag & mask) {
			PostMessage(pSharedData->hwnd[i], WM_APP+3, reinterpret_cast<WPARAM>(hwnd), CreateMessageArgAtom(wParam, lParam));
		}
	}
}

static void ProcessDestroyMessage(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	UINT uMsg;
	if(IsAfxWindow(hwnd)) {
		uMsg = WM_APP + 4;
	} else if(IsTApplication(hwnd)) {
		uMsg = WM_APP + 5;
	} else {
		return;
	}
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	for(int i = 0, mask = 1; i < MAX_HWND; ++i, mask <<= 1) {
		if(pSharedData->bFlag & mask) {
			SendMessage(pSharedData->hwnd[i], uMsg, reinterpret_cast<WPARAM>(hwnd), 0);
		}
	}
}

typedef void (*ProcessFunc)(HWND hwnd, WPARAM wParam, LPARAM lParam);
struct FuncTable
{
	UINT mes;
	bool check;
	ProcessFunc func;
} g_Table[] = {
	{ LB_GETCARETINDEX, false, ProcessGetCaretIndex },
	{ LB_SETCARETINDEX, false, ProcessSetCaretIndex },
	{ WM_MOVE, true, ProcessMoveMessage },
	{ WM_SIZE, false, ProcessSizeMessage },
	{ WM_ACTIVATE, true, ProcessActivateMessage },
	{ WM_DESTROY, false, ProcessDestroyMessage },
	{ 0, false, NULL }
};

LRESULT CALLBACK CallWndProc(int nCode, WPARAM wParam, LPARAM lParam)
{
	static bool fThrough = false; // for avoiding to execute twice
	LRESULT res;
	if(!fThrough) {
		CWPSTRUCT *pMessage = static_cast<CWPSTRUCT*>(reinterpret_cast<void*>(lParam));
		if(pMessage->message == g_uInjEnvMessage)      ProcessInjEnvMessage(pMessage->hwnd, pMessage->wParam, pMessage->lParam);
		else if(pMessage->message == g_uInjCmdMessage) ProcessInjCmdMessage(pMessage->hwnd, pMessage->wParam, pMessage->lParam);
		else if(pMessage->message == g_uInjMesMessage) ProcessInjMesMessage(pMessage->hwnd, pMessage->wParam, pMessage->lParam);
		else if(pMessage->message == g_uInjKeyMessage) ProcessInjKeyMessage(pMessage->hwnd, pMessage->wParam, pMessage->lParam);
		else {
			FuncTable *p = g_Table;
			while(p->func) {
				if((p->check ? IsAfxWindow(pMessage->hwnd) : true) && p->mes == pMessage->message) {
					(p->func)(pMessage->hwnd, pMessage->wParam, pMessage->lParam);
				}
				++p;
			}
		}
	}
	fThrough = true;
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	res = CallNextHookEx(pSharedData->hHook, nCode, wParam, lParam);
	fThrough = false;
	return res;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Call hook procedure for environment variable change
///
//////////////////////////////////////////////////////////////////////////////
static BOOL InjectEnvironmentVariableToA(HWND hwnd, LPSTR lpName, LPSTR lpValue)
{
	int nLenName = lstrlenA(lpName);
	int nLenValue = lstrlenA(lpValue);
	int nSize = nLenName + 1 + nLenValue + 1;
	SharedMem mem(KEY_INJENVVARARG, nSize * sizeof(CHAR));
	LPSTR pMem = static_cast<LPSTR>(mem.GetPointer());
	CopyMemory(pMem, lpName, sizeof(CHAR) * (nLenName + 1));
	CopyMemory(pMem + nLenName + 1, lpValue, sizeof(CHAR)* (nLenValue + 1));
	OutputDebugString("MessageSending for environment variable change");
	SendMessage(hwnd, g_uInjEnvMessage, nSize * sizeof(CHAR), 0);
	OutputDebugString("MessageSent for environment variable change");
	return *pMem == 0;
}

static BOOL InjectEnvironmentVariableToW(HWND hwnd, LPWSTR lpName, LPWSTR lpValue)
{
	int nLenName = lstrlenW(lpName);
	int nLenValue = lstrlenW(lpValue);
	int nSize = nLenName + 1 + nLenValue + 1;
	SharedMem mem(KEY_INJENVVARARG, nSize * sizeof(WCHAR));
	LPWSTR pMem = static_cast<LPWSTR>(mem.GetPointer());
	CopyMemory(pMem, lpName, sizeof(WCHAR) * (nLenName + 1));
	CopyMemory(pMem + nLenName + 1, lpValue, sizeof(WCHAR)* (nLenValue + 1));
	OutputDebugString("MessageSending for environment variable change");
	SendMessage(hwnd, g_uInjEnvMessage, nSize * sizeof(WCHAR), 1);
	OutputDebugString("MessageSent for environment variable change");
	return *pMem == 0;
}

extern "C" __declspec(dllexport) BOOL WINAPI InjectEnvironmentVariableA(HWND hwnd, LPSTR lpName, LPSTR lpValue)
{
	Str sName(lpName);
	Str sValue(lpValue);
	return IsWindowUnicode(hwnd)
		? InjectEnvironmentVariableToW(hwnd, sName, sValue)
		: InjectEnvironmentVariableToA(hwnd, sName, sValue);
}

extern "C" __declspec(dllexport) BOOL WINAPI InjectEnvironmentVariableW(HWND hwnd, LPWSTR lpName, LPWSTR lpValue)
{
	Str sName(lpName);
	Str sValue(lpValue);
	return IsWindowUnicode(hwnd)
		? InjectEnvironmentVariableToW(hwnd, sName, sValue)
		: InjectEnvironmentVariableToA(hwnd, sName, sValue);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Call hook procedure for command injection
///
//////////////////////////////////////////////////////////////////////////////
static BOOL InjectCommandToA(HWND hwnd, LPSTR lpCommand)
{
	int nLen = lstrlenA(lpCommand);
	int nSize = sizeof(int) + (nLen + 1) * sizeof(CHAR);
	SharedMem mem(KEY_INJCMDARG, nSize);
	LPSTR pMem = static_cast<LPSTR>(mem.GetPointer());
	CopyMemory(pMem, &nLen, sizeof(int));
	CopyMemory(pMem + sizeof(int), lpCommand, (nLen + 1) * sizeof(CHAR));
	OutputDebugString("MessageSending for command injection");
	SendMessage(hwnd, g_uInjCmdMessage, nSize, 0);
	OutputDebugString("MessageSent for command injection");
	return *pMem == 0;
}

static BOOL InjectCommandToW(HWND hwnd, LPWSTR lpCommand)
{
	int nLen = lstrlenW(lpCommand);
	OutputDebugStringW(lpCommand);
	int nSize = sizeof(PREFIX) + sizeof(int) + (nLen + 1) * sizeof(WCHAR);
	SharedMem mem(KEY_INJCMDARG, nSize);
	LPSTR pMem = static_cast<LPSTR>(mem.GetPointer());
	CopyMemory(pMem, PREFIX, sizeof(PREFIX));
	CopyMemory(pMem + sizeof(PREFIX), &nLen, sizeof(int));
	CopyMemory(pMem + sizeof(PREFIX) + sizeof(int), lpCommand, (nLen + 1) * sizeof(WCHAR));
	OutputDebugString("MessageSending for command injection");
	SendMessage(hwnd, g_uInjCmdMessage, nSize, 1);
	OutputDebugString("MessageSent for command injection");
	return *pMem == 0;
}

extern "C" __declspec(dllexport) BOOL WINAPI InjectCommandA(HWND hwnd, LPSTR lpCommand)
{
	Str sCommand(lpCommand);
	return IsWindowUnicode(hwnd)
		? InjectCommandToW(hwnd, sCommand)
		: InjectCommandToA(hwnd, sCommand);
}

extern "C" __declspec(dllexport) BOOL WINAPI InjectCommandW(HWND hwnd, LPWSTR lpCommand)
{
	Str sCommand(lpCommand);
	return IsWindowUnicode(hwnd)
		? InjectCommandToW(hwnd, sCommand)
		: InjectCommandToA(hwnd, sCommand);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Call hook procedure for message injection
///
//////////////////////////////////////////////////////////////////////////////
static BOOL InjectMessageToA(HWND hwnd, LPSTR lpMessage)
{
	int nLen = lstrlenA(lpMessage);
	int nSize = sizeof(int) * 2 + (nLen + 1) * sizeof(CHAR);
	SharedMem mem(KEY_INJMESARG, nSize);
	LPSTR pMem = static_cast<LPSTR>(mem.GetPointer());
	pMem[0] = pMem[1] = pMem[2] = pMem[3] = '\xFF';
	CopyMemory(pMem + sizeof(int), &nLen, sizeof(int));
	CopyMemory(pMem + sizeof(int) * 2, lpMessage, (nLen + 1) * sizeof(CHAR));
	OutputDebugString("MessageSending for message injection");
	SendMessage(hwnd, g_uInjMesMessage, nSize, 0);
	OutputDebugString("MessageSent for messaage injection");
	return *pMem == 0;
}

static BOOL InjectMessageToW(HWND hwnd, LPWSTR lpMessage)
{
	int nLen = lstrlenW(lpMessage);
	int nSize = sizeof(PREFIX) + sizeof(int) + (nLen + 1) * sizeof(WCHAR);
	SharedMem mem(KEY_INJMESARG, nSize);
	LPSTR pMem = static_cast<LPSTR>(mem.GetPointer());
	CopyMemory(pMem, PREFIX, sizeof(PREFIX));
	CopyMemory(pMem + sizeof(PREFIX), &nLen, sizeof(int));
	CopyMemory(pMem + sizeof(PREFIX) + sizeof(int), lpMessage, (nLen + 1) * sizeof(WCHAR));
	OutputDebugString("MessageSending for message injection");
	SendMessage(hwnd, g_uInjMesMessage, nSize, 1);
	OutputDebugString("MessageSent for messaage injection");
	return *pMem == 0;
}

extern "C" __declspec(dllexport) BOOL WINAPI InjectMessageA(HWND hwnd, LPSTR lpMessage)
{
	Str sMessage(lpMessage);
	return IsWindowUnicode(hwnd)
		? InjectMessageToW(hwnd, sMessage)
		: InjectMessageToA(hwnd, sMessage);
}

extern "C" __declspec(dllexport) BOOL WINAPI InjectMessageW(HWND hwnd, LPWSTR lpMessage)
{
	Str sMessage(lpMessage);
	return IsWindowUnicode(hwnd)
		? InjectMessageToW(hwnd, sMessage)
		: InjectMessageToA(hwnd, sMessage);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Call hook procedure for key injection
///
//////////////////////////////////////////////////////////////////////////////
extern "C" __declspec(dllexport) BOOL WINAPI InjectKey(HWND hwnd, int nKey)
{
	OutputDebugString("MessageSending for key injection");
	SendMessage(hwnd, g_uInjKeyMessage, nKey, 0);
	OutputDebugString("MessageSent for key injection");
	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Install hook procedure
///
//////////////////////////////////////////////////////////////////////////////
extern "C"__declspec(dllexport) BOOL WINAPI InstallInjectEnvHook(void)
{
	OutputDebugString("Installing hook");
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	pSharedData->hHook = SetWindowsHookEx(WH_CALLWNDPROC, (HOOKPROC)CallWndProc, g_hInstance, 0);
	return pSharedData->hHook != NULL;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Uninstall hook procedure
///
//////////////////////////////////////////////////////////////////////////////
extern "C" __declspec(dllexport) BOOL WINAPI UninstallInjectEnvHook(void)
{
	OutputDebugString("Uninstalling hook");
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	BOOL fResult = UnhookWindowsHookEx(pSharedData->hHook);
	pSharedData->hHook = NULL;
	return fResult;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Return boolean value if hook procedure is installed
///
//////////////////////////////////////////////////////////////////////////////
extern "C" __declspec(dllexport) BOOL WINAPI IsInjectEnvHookInstalled(void)
{
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	return pSharedData->hHook != NULL;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Register AfxOnMessage callback window
///
//////////////////////////////////////////////////////////////////////////////
extern "C" __declspec(dllexport) BOOL WINAPI RegisterAfxOnMessage(HWND hwnd)
{
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	for(int i = 0, mask = 1; i < MAX_HWND; ++i, mask <<= 1) {
		if(!(mask & pSharedData->bFlag)) {
			pSharedData->bFlag = pSharedData->bFlag | mask;
			pSharedData->hwnd[i] = hwnd;
			return TRUE;
		}
	}
	return FALSE;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Unregister AfxOnMessage callback window
///
//////////////////////////////////////////////////////////////////////////////
extern "C" __declspec(dllexport) BOOL WINAPI UnregisterAfxOnMessage(HWND hwnd)
{
	SharedData* pSharedData = static_cast<SharedData*>(g_pSharedMem->GetPointer());
	for(int i = 0, mask = 1; i < MAX_HWND; ++i, mask <<= 1) {
		if(mask & pSharedData->bFlag && pSharedData->hwnd[i] == hwnd) {
			pSharedData->bFlag = pSharedData->bFlag & ~mask;
			pSharedData->hwnd[i] = 0;
			return TRUE;
		}
	}
	return FALSE;
}

//////////////////////////////////////////////////////////////////////////////
///
/// DLL entry point
///
//////////////////////////////////////////////////////////////////////////////
extern "C" __declspec(dllexport) BOOL WINAPI DllMain(HINSTANCE hInst, DWORD dwReason, LPVOID)
{
	switch(dwReason) {
	case DLL_PROCESS_ATTACH:
		g_pSharedMem = new SharedMem(KEY_INJENVVARCTX, sizeof(SharedData));
		g_hInstance = hInst;
		g_uInjEnvMessage = RegisterWindowMessage(KEY_INJENVVARMES);
		g_uInjCmdMessage = RegisterWindowMessage(KEY_INJCMDMES);
		g_uInjMesMessage = RegisterWindowMessage(KEY_INJMESMES);
		g_uInjKeyMessage = RegisterWindowMessage(KEY_INJKEYMES);
		break;
	case DLL_PROCESS_DETACH:
		delete g_pSharedMem;
		break;
	}
	return TRUE;
}
